# Copyright (c) 2013 Ian C. Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import os
import os.path
from ConfigParser import SafeConfigParser, NoSectionError, NoOptionError

from .exceptions import MygrateError
from .callbacks import MygrateCallbacks


class MygrateConfigError(MygrateError):
    """Informational exception for bad MyGrate configuration."""
    pass


class MygrateConfig(object):

    def __init__(self):
        self.parser = SafeConfigParser()

        fnames = self._get_paths()
        if not self.parser.read(fnames):
            raise MygrateConfigError('No valid config files in '+repr(fnames))

    def _get_paths(self):
        env_config = os.getenv('MYGRATE_CONFIG', None)
        if env_config:
            return [os.path.expanduser(env_config)]
        else:
            home_config = os.path.expanduser('~/.mygrate.conf')
            global_config = '/etc/mygrate.conf'
            return [home_config, global_config]

    def get_callbacks(self):
        try:
            section = self.parser.options('callbacks')
        except (NoSectionError):
            msg = 'Configuration section `callbacks` required.'
            raise MygrateConfigError(msg)
        ret = MygrateCallbacks()
        for table in section:
            mod_name = self.parser.get('callbacks', table)
            mod = __import__(mod_name)
            ret.register_module(table, mod)
        return ret

    def get_mysql_connection_info(self):
        ret = {}
        mapping = {'host': 'host',
                   'port': 'port',
                   'user': 'user',
                   'passwd': 'password',
                   'unix_socket': 'unix_socket'}
        for key, value in mapping.items():
            try:
                ret[key] = self.parser.get('mysql', value)
            except (NoSectionError, NoOptionError):
                pass
        return ret

    def get_mysql_binlog_info(self):
        try:
            index_file = self.parser.get('binlog', 'index_file')
        except (NoSectionError, NoOptionError):
            index_file = '/var/lib/mysql/mysql-bin.index'
        if not os.path.exists(index_file):
            raise MygrateConfigError('Invalid binlog index file: '+index_file)
        try:
            delay = self.parser.getfloat('binlog', 'tracking_delay')
        except (NoSectionError, NoOptionError):
            delay = 1.0
        return index_file, float(delay)

    def get_tracking_dir(self):
        try:
            tracking_dir = self.parser.get('binlog', 'tracking_dir')
        except (NoSectionError, NoOptionError):
            tracking_dir = os.path.expanduser('~/.binlog-tracking')
            try:
                os.makedirs(tracking_dir)
            except OSError:
                pass
        tracking_dir = os.path.expanduser(tracking_dir)
        if not os.path.isdir(tracking_dir):
            msg = 'Tracking directory does not exist: '+tracking_dir
            raise MygrateConfigError(msg)
        return tracking_dir


cfg = MygrateConfig()


# vim:et:fdm=marker:sts=4:sw=4:ts=4



# Copyright (c) 2013 Ian C. Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#


import os
import os.path
import sys
import optparse
import cPickle
import MySQLdb
import MySQLdb.cursors


class InitialQuery(object):
    """Manages direct queries to MySQL for importing, validation, and
    re-validation.

    """

    def __init__(self, mysql_info, callbacks, action='INSERT'):
        self.mysql_info = mysql_info
        self.callbacks = callbacks
        self.action = action

    def run_callback(self, table, cols):
        """Executes the INSERT callback for the given table.

        :param table: The table being imported from.
        :param cols: The column dict containing the row data.

        """
        self.callbacks.execute(table, self.action, cols)

    def get_connection(self, db, streaming=True):
        """Creates and returns a connection to the MySQL server and selects the
        given database.

        :param db: The database name to select after connection.
        :param streaming: Whether or not a streaming MySQL cursor should be
                          used, for large data sets.
        :returns: A MySQL connection object.

        """
        kwargs = self.mysql_info.copy()
        kwargs['db'] = db
        kwargs['charset'] = 'utf8'
        if streaming:
            kwargs['cursorclass'] = MySQLdb.cursors.SSDictCursor
        else:
            kwargs['cursorclass'] = MySQLdb.cursors.DictCursor
        conn = MySQLdb.connect(**kwargs)
        return conn

    def process_table(self, full_table):
        """Does a streaming, full-table SELECT on the requested table to
        retrieve every row. Rows are then passed to `run_callback()`.

        :param full_table: The database and table names, separated by a period,
                           as given on the command line.

        """
        db, table = full_table.split('.')
        conn = self.get_connection(db)
        cur = conn.cursor()
        try:
            sql = 'SELECT * FROM `{0}`'.format(table)
            cur.execute(sql)
            while True:
                rows = cur.fetchmany(1000)
                if not rows:
                    break
                for row in rows:
                    self.run_callback(full_table, row)
        finally:
            cur.close()
            conn.close()


def main():
    from .config import cfg

    description = """\
This program attempts to import an entire table by creating job tasks for each
row.

Configuration for %prog is done with configuration files. This is either
/etc/mygrate.conf, ~/.mygrate.conf, or an alternative specified by the
MYGRATE_CONFIG environment variable.
"""
    usage = 'usage: %prog [options] <database>.<table> ...'
    op = optparse.OptionParser(usage=usage, description=description)
    options, requested_tables = op.parse_args()
    if not requested_tables:
        op.error('Must supply at least one database and table.')

    mysql_info = cfg.get_mysql_connection_info()
    callbacks = cfg.get_callbacks()

    query = InitialQuery(mysql_info, callbacks)
    for table in requested_tables:
        query.process_table(table)


if __name__ == '__main__':
    main()


# vim:et:fdm=marker:sts=4:sw=4:ts=4




import os.path
import cPickle
import MySQLdb
import MySQLdb.cursors

from mox import MoxTestBase, IsA

from mygrate.query import InitialQuery


class TestInitialQuery(MoxTestBase):

    def test_run_callback(self):
        callbacks = self.mox.CreateMockAnything()
        callbacks.execute('test.test', 'TEST', {'one': 1, 'two': 2})
        self.mox.ReplayAll()
        importer = InitialQuery(None, callbacks, 'TEST')
        importer.run_callback('test.test', {'one': 1, 'two': 2})

    def test_get_connect(self):
        self.mox.StubOutWithMock(MySQLdb, 'connect')
        MySQLdb.connect(host='testhost', user='testuser', passwd='testpass',
                        db='testdb', charset='utf8', cursorclass=MySQLdb.cursors.SSDictCursor).AndReturn(13)
        options = {'host': 'testhost', 'user': 'testuser', 'passwd': 'testpass'}
        self.mox.ReplayAll()
        importer = InitialQuery(options, None, None)
        self.assertEqual(13, importer.get_connection('testdb'))

    def test_process_table(self):
        importer = InitialQuery(None, None, None)
        self.mox.StubOutWithMock(importer, 'get_connection')
        self.mox.StubOutWithMock(importer, 'run_callback')
        conn = self.mox.CreateMockAnything()
        importer.get_connection('testdb').AndReturn(conn)
        cur = self.mox.CreateMockAnything()
        conn.cursor().AndReturn(cur)
        cur.execute('SELECT * FROM `testtable`')
        cur.fetchmany(IsA(int)).AndReturn([{'one': 1, 'two': 2}, {'one': 3, 'two': 4}])
        importer.run_callback('testdb.testtable', {'one': 1, 'two': 2})
        importer.run_callback('testdb.testtable', {'one': 3, 'two': 4})
        cur.fetchmany(IsA(int)).AndReturn([])
        cur.close()
        conn.close()
        self.mox.ReplayAll()
        importer.process_table('testdb.testtable')


# vim:et:fdm=marker:sts=4:sw=4:ts=4



# Copyright (c) 2013 Ian C. Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#


import os
import os.path
import sys
import optparse
import cPickle
import MySQLdb
import MySQLdb.cursors


class MygrateCallbacks(object):
    """Manages registration of callbacks for actions against tables.

    """

    def __init__(self):
        self.callbacks = {}

    def get_registered_tables(self):
        """Returns a list of tables for which callbacks have been registered.

        """
        return self.callbacks.keys()

    def register(self, table, action, callback):
        """Registers a callback for a single action on a given table.

        :param table: The table the callback should apply to.
        :param action: The action the callback should apply to.
        :param callback: The function to call when the action happens on the
                         table.

        """
        self.callbacks.setdefault(table, {})
        self.callbacks[table][action] = callback

    def register_module(self, table, module):
        """Registers a module containing all the action callbacks for a table.

        :param table: The table the callbacks should apply to.
        :param module: A module (or other object) containing "INSERT",
                       "UPDATE", or "DELETE" attributes that correspond to the
                       callbacks that should be executed.

        """
        self.callbacks.setdefault(table, {})
        if hasattr(module, 'INSERT'):
            self.callbacks[table]['INSERT'] = module.INSERT
        if hasattr(module, 'UPDATE'):
            self.callbacks[table]['UPDATE'] = module.UPDATE
        if hasattr(module, 'DELETE'):
            self.callbacks[table]['DELETE'] = module.DELETE

    def execute(self, table, action, *args, **kwargs):
        """Executes the callback registered to the given table and action. If
        no callback has been registered, nothing is executed.

        :param table: The table to execute the callback for. This table is
                      passed in as the first positional argument to the
                      callback.
        :param action: The action to execute the callback for.
        :param args: The positional arguments to pass in to the callback.
        :param kwargs: The keyword arguments to pass in to the callback.

        """
        if table not in self.callbacks:
            return
        if action not in self.callbacks[table]:
            return
        callback = self.callbacks[table][action]
        callback(table, *args, **kwargs)


# vim:et:fdm=marker:sts=4:sw=4:ts=4




import os
import sys
import os.path
import re
import signal
import subprocess
import optparse
import traceback
import glob
import time
from datetime import datetime
from ast import literal_eval
from time import sleep

import MySQLdb
import bitstring


class ValueParser(object):
    """The mysqlbinlog command has its own unique way of serializing its
    various data types to standard output. This class tries several different
    methods that should cover almost all data types.

    The string formats were obtained by examining sql/log_event.cc from
    Community MySQL Server version 5.6.10.

    """

    __slots__ = ['line', 'valid', 'value']
    _int_pattern = re.compile(r'^[-0-9]+ \(([0-9]+)\)$')

    def __init__(self, line):
        self.line = line
        self.valid = False
        self.value = None
        try:
            self._try_null()
            self._try_quoted()
            self._try_int()
            self._try_float()
            self._try_datetime()
            self._try_time()
            self._try_date()
            self._try_bitstring()
        except StopIteration:
            self.valid = True

    def _literal_eval(self, data):
        try:
            self.value = literal_eval(data)
        except Exception:
            pass
        else:
            raise StopIteration()

    def _try_null(self):
        if self.line == 'NULL':
            raise StopIteration()

    def _try_quoted(self):
        if self.line[0] != "'" or self.line[-1] != "'":
            return
        data = self.line[1:-1].encode('string_escape')
        self._literal_eval("'{0}'".format(data))

    def _try_int(self):
        match = self._int_pattern.match(self.line)
        if not match:
            return
        self._literal_eval(match.group(1))

    def _try_float(self):
        try:
            self.value = float(self.line)
        except ValueError:
            pass
        else:
            raise StopIteration()

    def _try_datetime(self):
        try:
            self.value = datetime.strptime(self.line, '%Y-%m-%d %H:%M:%S')
        except ValueError:
            pass
        else:
            raise StopIteration()

    def _try_date(self):
        try:
            self.value = datetime.strptime(self.line, '%Y-%m-%d')
        except ValueError:
            pass
        else:
            raise StopIteration()

    def _try_time(self):
        try:
            self.value = time.strptime(self.line, '%H:%M:%S')
        except ValueError:
            pass
        else:
            raise StopIteration()

    def _try_bitstring(self):
        if self.line[:2] != "b'" or self.line[-1] != "'":
            return
        data = self.line[2:-1]
        try:
            bitstr = bitstring.Bits(bin='0b{0}'.format(data))
        except bitstring.Error:
            pass
        else:
            self.value = bitstr.tobytes()
            raise StopIteration()

    @classmethod
    def parse(cls, line):
        return cls(line).value


class QueryBase(object):
    """The base class for keeping track of a single query loaded in from the
    command:

        mysqlbinlog -v --base64-output=DECODE-ROWS /path/to/binlog.00000x

    The binlog must be in ROW format. This format is easier for a machine to
    process.

    """

    _column_pattern = re.compile(r'^  @(\d+)=(.*)$')
    _int_pattern = re.compile(r'\(([^\)]*)\)')

    def __init__(self, line, callbacks, column_names, char_sets):
        self.callbacks = callbacks
        self.column_names = column_names
        self.char_sets = char_sets
        self.table = None
        self.invalid = False
        self.current_value_type = None
        self.values = {'WHERE': [], 'SET': []}
        self._parse_initial_line(line)

    def _parse_initial_line(self, line):
        match = self._initial_pattern.match(line)
        if not match:
            self.invalid = True
            return
        identifiers = [ident.strip('`') for ident in match.group(1).split('.')]
        self.table = '.'.join(identifiers)

    def _parse_value(self, value, char_set):
        parsed = ValueParser.parse(value)
        if char_set and isinstance(parsed, str):
            return parsed.decode(char_set)
        return parsed

    def parse(self, line):
        """Adds the contents of the given line to the current query. The
        mysqlbinlog command uses two basic sections `SET` and `WHERE` to define
        the different parts of the queries.

        :param line: The line to parse into the query.

        """
        if line == 'SET' or line == 'WHERE':
            self.current_value_type = line
            return
        match = self._column_pattern.match(line)
        if not match:
            self.invalid = True
            return
        char_set = self.char_sets.get(self.table)
        col_value = self._parse_value(match.group(2), char_set)
        self.values[self.current_value_type].append(col_value)

    def __repr__(self):
        return '<Query {0} {1} WHERE={2!s} SET={3!s}>'.format(
            self.type,
            self.table,
            self.values['WHERE'],
            self.values['SET'])


class InsertQuery(QueryBase):
    _initial_pattern = re.compile(r'^INSERT INTO (.*)$')
    type = 'INSERT'

    def finish(self):
        """When the query is finished, its numeric column references are
        translated into column names using a lookup table, resulting in a dict
        object that is then passed to the callback.

        """
        table_column_names = self.column_names[self.table]
        set_vals = {}
        for i, val in enumerate(self.values['SET']):
            key = table_column_names[i]
            set_vals[key] = val
        self.callbacks.execute(self.table, self.type, set_vals)


class UpdateQuery(QueryBase):
    _initial_pattern = re.compile(r'^UPDATE (.*)$')
    type = 'UPDATE'

    def finish(self):
        """When the query is finished, its numeric column references are
        translated into column names using a lookup table, resulting in a dict
        object that is then passed to the callback.

        """
        table_column_names = self.column_names[self.table]
        where_vals = {}
        for i, val in enumerate(self.values['WHERE']):
            key = table_column_names[i]
            where_vals[key] = val
        set_vals = {}
        for i, val in enumerate(self.values['SET']):
            key = table_column_names[i]
            set_vals[key] = val
        self.callbacks.execute(self.table, self.type, where_vals, set_vals)


class DeleteQuery(QueryBase):
    _initial_pattern = re.compile(r'^DELETE FROM (.*)$')
    type = 'DELETE'

    def finish(self):
        """When the query is finished, its numeric column references are
        translated into column names using a lookup table, resulting in a dict
        object that is then passed to the callback.

        """
        table_column_names = self.column_names[self.table]
        where_vals = {}
        for i, val in enumerate(self.values['WHERE']):
            key = table_column_names[i]
            where_vals[key] = val
        self.callbacks.execute(self.table, self.type, where_vals)


class QueryParser(object):
    """Keeps track of the current query being read from the binlog. If it sees a
    new query, it marks the previous query as finished and sends it to the
    callback.

    """

    def __init__(self, callbacks, column_names, char_sets):
        self.current = None
        self.callbacks = callbacks
        self.column_names = column_names
        self.char_sets = char_sets

    def parse(self, line):
        """Checks if the line is the beginning of a new query or should be added
        on to the current query.

        :param line: The line to process.

        """
        if line.startswith('INSERT'):
            self._handle_completion()
            self.current = InsertQuery(line, self.callbacks, self.column_names,
                                       self.char_sets)
        elif line.startswith('UPDATE'):
            self._handle_completion()
            self.current = UpdateQuery(line, self.callbacks, self.column_names,
                                       self.char_sets)
        elif line.startswith('DELETE'):
            self._handle_completion()
            self.current = DeleteQuery(line, self.callbacks, self.column_names,
                                       self.char_sets)
        elif self.current:
            self.current.parse(line)

    def finish(self):
        """Called at the end of the binlog, so that the current query can be
        marked complete and sent to the callback.

        """
        self._handle_completion()
        self.current = None

    def _handle_completion(self):
        if self.current:
            self.current.finish()


class BinlogParser(object):

    def __init__(self, index_file, pos_dir, callbacks, column_names=None,
                 char_sets=None):
        self.done = False
        self.index_file = index_file
        self.pos_dir = pos_dir
        self.callbacks = callbacks
        self.column_names = column_names or {}
        self.char_sets = char_sets or {}
        self.binlog_mtimes = {}

    def _load_one_table_names(self, conn, db, table):
        cur = conn.cursor()
        try:
            cur.execute("""SELECT `COLUMN_NAME` FROM
                           `INFORMATION_SCHEMA`.`COLUMNS`
                           WHERE `TABLE_SCHEMA`=%s AND `TABLE_NAME`=%s""",
                        (db, table))
            return [row[0] for row in cur.fetchall()]
        finally:
            cur.close()

    def load_column_names(self, mysql_info):
        """Connects to the MySQL server and loads column names for all the
        tables for which there are callbacks.

        :param mysql_info: Contains the details about the MySQL connection.

        """
        kwargs = mysql_info.copy()
        conn = MySQLdb.connect(**kwargs)

        try:
            for full_table in self.callbacks.get_registered_tables():
                db, table = full_table.split('.', 1)
                names = self._load_one_table_names(conn, db, table)
                self.column_names[full_table] = names
        finally:
            conn.close()

    def _load_one_table_charset(self, conn, db, table):
        cur = conn.cursor()
        try:
            cur.execute("""SELECT `CCSA`.`CHARACTER_SET_NAME` FROM
`INFORMATION_SCHEMA`.`TABLES` `T`,
`INFORMATION_SCHEMA`.`COLLATION_CHARACTER_SET_APPLICABILITY` `CCSA`
WHERE `CCSA`.`COLLATION_NAME` = `T`.`TABLE_COLLATION`
AND `T`.`TABLE_SCHEMA` = %s
AND `T`.`TABLE_NAME` = %s""", (db, table))
            row = cur.fetchone()
            return row[0] if row else None
        finally:
            cur.close()

    def load_character_sets(self, mysql_info):
        """Connects to the MySQL server and loads table character sets for all
        the tables for which there are callbacks.

        :param mysql_info: Contains the details about the MySQL connection.

        """
        kwargs = mysql_info.copy()
        conn = MySQLdb.connect(**kwargs)

        try:
            for full_table in self.callbacks.get_registered_tables():
                db, table = full_table.split('.', 1)
                charset = self._load_one_table_charset(conn, db, table)
                self.char_sets[full_table] = charset
        finally:
            conn.close()

    def read_position(self, pos_file):
        """Reads the latest binlog position from the position tracking file.

        :param pos_file: The path to the position tracking file.
        :returns: The position from the file, if it exists, or 0.

        """
        try:
            with open(pos_file, 'r') as f:
                ret = f.read().rstrip()
                if ret:
                    return ret
        except IOError, (err, s):
            if err != 2:
                raise
        return '0'

    def write_position(self, f, pos):
        """Writes the binlog position to the position tracking file, immediately
        flushing write caches to disk. The file is always truncated before
        being written to so that its contents are only the latest position.

        :param f: The file object to write to.
        :param pos: The position to write to f.

        """
        f.seek(0, os.SEEK_SET)
        f.truncate()
        f.write(pos)
        f.flush()

    def read_index(self):
        """Reads the known binlog file paths from the binlog index file, which
        is managed by MySQL.

        :returns: List of file paths to the binlogs.

        """
        index_dir = os.path.dirname(self.index_file)
        with open(self.index_file, 'r') as f:
            ret = []
            for line in f:
                binlog = line.rstrip()
                binlog = os.path.normpath(os.path.join(index_dir, binlog))
                ret.append(binlog)
            return ret

    def build_pos_file(self, binlog):
        """Given a binlog file path, build a corresponding position tracking
        file path based on the configured tracking directory.

        :param binlog: The binlog file path.
        :returns: The corresponding path to the tracking file.

        """
        binlog_base, binlog_ext = os.path.splitext(binlog)
        return os.path.join(self.pos_dir, 'binlogpos'+binlog_ext)

    def process_binlog(self, binlog):
        """Sweeps through a single binlog, checking it for updates after the
        last known position read from the tracking file. If new positions are
        seen, the tracking file is updated with the new position. Queries are
        processed on the spot and sent to the callback.

        :param binlog: The path to the binlog file.

        """
        p = QueryParser(self.callbacks, self.column_names, self.char_sets)

        pos_file = self.build_pos_file(binlog)

        last_position = self.read_position(pos_file)
        print 'processing', binlog, 'from', last_position
        writepos = open(pos_file, 'w')
        self.write_position(writepos, last_position)

        args = ['mysqlbinlog', '-v', '--base64-output=DECODE-ROWS', binlog,
                '-j', last_position,
                '--set-charset=utf8']

        proc = subprocess.Popen(args, stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE)
        proc.stdin.close()

        try:
            for line in proc.stdout:
                if self.done:
                    break
                if line.startswith('### '):
                    p.parse(line[4:].rstrip('\r\n'))
                elif line.startswith('# at '):
                    last_position = line[5:].rstrip()
                    self.write_position(writepos, last_position)
            else:
                p.finish()
        except Exception:
            traceback.print_exc()
        finally:
            writepos.close()
            proc.wait()

    def process_all_binlogs(self):
        """Sweeps through all the binlogs in the index. The index is read every
        sweep in case MySQL is restarted or rotates to a new binlog file. This
        method also checks each binlogs mtime to see if it has been modified
        since the last sweep.

        """
        binlogs = self.read_index()
        for binlog in binlogs:
            if not self.done:
                old_mtime = self.binlog_mtimes.get(binlog, 0.0)
                self.binlog_mtimes[binlog] = float(os.path.getmtime(binlog))
                if old_mtime < self.binlog_mtimes[binlog]:
                    self.process_binlog(binlog)

    def set_binlogpos_at_end(self, binlog):
        """Manually sets the binlog's position tracking file to the end of the
        binlog, so that future sweeps will not act upon any existing entries.

        :param binlog: The file path to the binlog.

        """
        pos_file = self.build_pos_file(binlog)

        old_pos = self.read_position(pos_file)
        binlog_size = os.path.getsize(binlog)
        with open(pos_file, 'w') as f:
            self.write_position(f, str(binlog_size))
        print 'changing', pos_file, 'from', old_pos, 'to', binlog_size


def confirm_skip_existing():
    print 'This utility will seek the binlog tracking files to the end of all'
    print 'existing entries. All previous entries will be skipped. There is no'
    print 'easy way to undo this operation!'
    print
    while True:
        answer = raw_input('Are you sure?  N/y: ')
        if answer in ['y', 'Y']:
            return
        elif answer in ['n', 'N', '']:
            sys.exit(1)


def skip_existing():
    """This function is declared as the entry point for the
    `mygrate-skip` command.

    """
    from .config import cfg

    description = """\
This program calculates the latest binlog positions and creates/modifies the
tracking files to those positions. Subsequent executions of the binlog parser
will start at these new positions.

Configuration for %prog is done with configuration files. This is either
/etc/mygrate.ini, ~/.mygrate.ini, or an alternative specified by the
MYGRATE_CONFIG environment variable.
"""
    op = optparse.OptionParser()
    op.add_option('-f', '--force', action='store_true', default=False,
                  help='Do not ask for confirmation, just do it.')
    options, extra = op.parse_args()

    if not options.force:
        confirm_skip_existing()

    binlog_index, tracking_delay = cfg.get_mysql_binlog_info()
    tracking_dir = cfg.get_tracking_dir()
    callbacks = cfg.get_callbacks()
    parser = BinlogParser(binlog_index, tracking_dir, callbacks)
    binlogs = parser.read_index()
    for binlog in binlogs:
        parser.set_binlogpos_at_end(binlog)


def main():
    """This function is declared the entry point for the
    `mygrate-binlog` command.

    """
    from .config import cfg

    description = """\
This program follows changes in the MySQL binlog, producing job tasks for each
change. It is intended to be long-running, and will briefly pause after
catching up each binlog before checking for new changes.

Configuration for %prog is done with configuration files. This is either
/etc/mygrate.ini, ~/.mygrate.ini, or an alternative specified by the
MYGRATE_CONFIG environment variable.
"""
    op = optparse.OptionParser(description=description)
    op.parse_args()

    tracking_dir = cfg.get_tracking_dir()
    binlog_index, tracking_delay = cfg.get_mysql_binlog_info()
    callbacks = cfg.get_callbacks()
    parser = BinlogParser(binlog_index, tracking_dir, callbacks)

    def graceful_quit(sig, frame):
        parser.done = True

    signal.signal(signal.SIGINT, graceful_quit)
    signal.signal(signal.SIGTERM, graceful_quit)

    mysql_info = cfg.get_mysql_connection_info()
    parser.load_column_names(mysql_info)
    parser.load_character_sets(mysql_info)

    while not parser.done:
        parser.process_all_binlogs()
        if not parser.done:
            sleep(tracking_delay)


if __name__ == '__main__':
    main()


# vim:et:fdm=marker:sts=4:sw=4:ts=4



# Copyright (c) 2013 Ian C. Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from setuptools import setup, find_packages


setup(name='mygrate',
      version='0.2',
      author='Ian Good',
      author_email='ian.good@rackspace.com',
      description='Facilitates the migration of data from MySQL or MariaDB with no downtime.',
      license='MIT',
      url='https://github.com/icgood/mygrate',
      packages=find_packages(),
      entry_points={'console_scripts': [
              'mygrate-query = mygrate.query:main',
              'mygrate-binlog = mygrate.binlog:main',
              'mygrate-skip = mygrate.binlog:skip_existing'
          ]},
      install_requires=['MySQL-python',
                        'bitstring'],
      classifiers=['Development Status :: 3 - Alpha',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Information Technology',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python'])


# vim:et:fdm=marker:sts=4:sw=4:ts=4




import os
import os.path
import tempfile
import shutil
import subprocess
from datetime import datetime

from mox import MoxTestBase, IsA

from mygrate.binlog import (QueryBase, InsertQuery, UpdateQuery,
                            DeleteQuery, QueryParser, BinlogParser)


class TestQueryParsing(MoxTestBase):

    def test_insertquery_parse_initial_line(self):
        q = InsertQuery('INSERT INTO `testdb`.`testtable`', None, None, None)
        self.assertEqual('testdb.testtable', q.table)

    def test_updatequery_parse_initial_line(self):
        q = UpdateQuery('UPDATE `testdb`.`testtable`', None, None, None)
        self.assertEqual('testdb.testtable', q.table)

    def test_deletequery_parse_initial_line(self):
        q = DeleteQuery('DELETE FROM `testdb`.`testtable`', None, None, None)
        self.assertEqual('testdb.testtable', q.table)

    def test_querybase_parse_value(self):
        q = InsertQuery('INSERT INTO `testdb`.`testtable`', None, None, {})
        self.assertEqual('as\'df', q._parse_value("'as'df'", None))
        self.assertEqual('as\\x00df', q._parse_value("'as\\x00df'", None))
        self.assertEqual(9, q._parse_value('9', None))
        self.assertEqual(None, q._parse_value('NULL', None))
        self.assertEqual(datetime(2013, 1, 1, 13, 30), q._parse_value('2013-01-01 13:30:00', None))
        self.assertEqual(65531, q._parse_value('-5 (65531)', None))
        self.assertEqual('test', q._parse_value("b'01110100011001010111001101110100'", None))
        self.assertEqual(None, q._parse_value('invalid', None))

    def test_querybase_parse_value_encoding(self):
        q = InsertQuery('INSERT INTO `testdb`.`testtable`', None, None, {})
        self.assertEqual(u'asdf', q._parse_value("'asdf'", 'latin1'))
        self.assertEqual('10\xc3\xb72=Five', q._parse_value("'10\xf72=Five'", 'latin1').encode('utf8'))

    def test_querybase_parse(self):
        q = UpdateQuery('UPDATE `testdb`.`testtable`', None, None, {})
        q.parse("SET")
        q.parse("  @1='asdf'")
        q.parse("  @2=9")
        q.parse("WHERE")
        q.parse("  @1='jkl'")
        q.parse("  @2=10")
        self.assertEqual(['asdf', 9], q.values['SET'])
        self.assertEqual(['jkl', 10], q.values['WHERE'])

    def test_insertquery_finish(self):
        callbacks = self.mox.CreateMockAnything()
        callbacks.execute('testdb.testtable', 'INSERT', {'one': 'asdf', 'two': None})
        self.mox.ReplayAll()
        q = InsertQuery('INSERT INTO `testdb`.`testtable`', callbacks, {'testdb.testtable': ['one', 'two']}, {})
        q.parse("SET")
        q.parse("  @1='asdf'")
        q.parse("  @2=NULL")
        q.finish()

    def test_updatequery_finish(self):
        callbacks = self.mox.CreateMockAnything()
        callbacks.execute('testdb.testtable', 'UPDATE', {'one': 'jkl', 'two': None}, {'one': 'asdf', 'two': None})
        self.mox.ReplayAll()
        q = UpdateQuery('UPDATE `testdb`.`testtable`', callbacks, {'testdb.testtable': ['one', 'two']}, {})
        q.parse("SET")
        q.parse("  @1='asdf'")
        q.parse("  @2=NULL")
        q.parse("WHERE")
        q.parse("  @1='jkl'")
        q.parse("  @2=NULL")
        q.finish()

    def test_deletequery_finish(self):
        callbacks = self.mox.CreateMockAnything()
        callbacks.execute('testdb.testtable', 'DELETE', {'one': 'asdf', 'two': None})
        self.mox.ReplayAll()
        q = DeleteQuery('DELETE FROM `testdb`.`testtable`', callbacks, {'testdb.testtable': ['one', 'two']}, {})
        q.parse("WHERE")
        q.parse("  @1='asdf'")
        q.parse("  @2=NULL")
        q.finish()

    def test_queryparser(self):
        callbacks = self.mox.CreateMockAnything()
        callbacks.execute('testdb.testtable', 'INSERT', {'one': 'asdf', 'two': None})
        callbacks.execute('testdb.testtable', 'UPDATE', {'one': 'jkl', 'two': None}, {'one': 'asdf', 'two': None})
        callbacks.execute('testdb.testtable', 'DELETE', {'one': 'asdf', 'two': None})
        self.mox.ReplayAll()
        qp = QueryParser(callbacks, {'testdb.testtable': ['one', 'two']}, {})
        qp.parse("INSERT INTO `testdb`.`testtable`")
        qp.parse("SET")
        qp.parse("  @1='asdf'")
        qp.parse("  @2=NULL")
        qp.parse("UPDATE `testdb`.`testtable`")
        qp.parse("SET")
        qp.parse("  @1='asdf'")
        qp.parse("  @2=NULL")
        qp.parse("WHERE")
        qp.parse("  @1='jkl'")
        qp.parse("  @2=NULL")
        qp.parse("DELETE FROM `testdb`.`testtable`")
        qp.parse("WHERE")
        qp.parse("  @1='asdf'")
        qp.parse("  @2=NULL")
        qp.finish()
        self.assertEqual(None, qp.current)


class TestBinlogParser(MoxTestBase):

    def setUp(self):
        super(TestBinlogParser, self).setUp()
        self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        super(TestBinlogParser, self).tearDown()
        shutil.rmtree(self.tmp_dir)

    def test_read_position(self):
        blp = BinlogParser(None, None, None, None)
        self.assertEqual('0', blp.read_position('file does not exist'))
        with tempfile.NamedTemporaryFile() as f:
            f.write('1234')
            f.flush()
            os.fsync(f.fileno())
            self.assertEqual('1234', blp.read_position(f.name))

    def test_write_position(self):
        blp = BinlogParser(None, None, None, None)
        with tempfile.NamedTemporaryFile() as f:
            self.assertEqual('0', blp.read_position(f.name))
            blp.write_position(f, '4321')
            self.assertEqual('4321', blp.read_position(f.name))
            blp.write_position(f, '1337')
            self.assertEqual('1337', blp.read_position(f.name))

    def test_read_index(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write('/path/to/file.1\n/path/to/file.2\n/path/to/file.3\n')
            f.flush()
            os.fsync(f.fileno())
            blp = BinlogParser(f.name, None, None, None)
            expected_binlogs = ['/path/to/file.1',
                                '/path/to/file.2',
                                '/path/to/file.3']
            self.assertEqual(expected_binlogs, blp.read_index())

    def test_build_pos_file(self):
        blp = BinlogParser(None, '/path/to/posdir', None, None)
        self.assertEqual('/path/to/posdir/binlogpos.000005', blp.build_pos_file('/path/to/mysql-binlog.000005'))
        self.assertEqual('/path/to/posdir/binlogpos.000001', blp.build_pos_file('/test/test.000001'))

    def test_process_binlog(self):
        pos_file = os.path.join(self.tmp_dir, 'binlogpos.000001')
        with open(pos_file, 'w') as f:
            f.write('1234')
        self.mox.StubOutWithMock(subprocess, 'Popen')
        proc = self.mox.CreateMockAnything()
        proc.stdin = self.mox.CreateMockAnything()
        proc.stdout = ['### INSERT INTO `testdb`.`testtable`\n',
                       '### SET\n',
                       "###   @1='asdf'\n",
                       "###   @2=NULL\n",
                       "# at 4321\n"]
        subprocess.Popen(['mysqlbinlog', '-v', '--base64-output=DECODE-ROWS', '/path/to/binlog.000001', '-j', '1234', '--set-charset=utf8'],
                         stdin=subprocess.PIPE, stdout=subprocess.PIPE).AndReturn(proc)
        proc.stdin.close()
        callbacks = self.mox.CreateMockAnything()
        callbacks.execute('testdb.testtable', 'INSERT', {'one': 'asdf', 'two': None})
        proc.wait()
        self.mox.ReplayAll()
        blp = BinlogParser(None, self.tmp_dir, callbacks, {'testdb.testtable': ['one', 'two']})
        blp.process_binlog('/path/to/binlog.000001')
        self.assertEqual('4321', blp.read_position(pos_file))

    def test_process_all_binlogs(self):
        blp = BinlogParser(None, None, None, None)
        self.mox.StubOutWithMock(blp, 'read_index')
        self.mox.StubOutWithMock(blp, 'process_binlog')
        self.mox.StubOutWithMock(os.path, 'getmtime')
        blp.read_index().AndReturn(['/path/to/test.1', '/path/to/test.2'])
        os.path.getmtime('/path/to/test.1').AndReturn(10.0)
        blp.process_binlog('/path/to/test.1')
        os.path.getmtime('/path/to/test.2').AndReturn(20.0)
        blp.process_binlog('/path/to/test.2')
        self.mox.ReplayAll()
        blp.process_all_binlogs()

    def test_set_binlogpos_at_end(self):
        binlog_file = os.path.join(self.tmp_dir, 'binlog.001')
        blp = BinlogParser(None, self.tmp_dir, None, None)
        with open(binlog_file, 'w') as f:
            f.write('x'*256)
        blp.set_binlogpos_at_end(binlog_file)
        self.assertEqual('256', blp.read_position(os.path.join(self.tmp_dir, 'binlogpos.001')))


# vim:et:fdm=marker:sts=4:sw=4:ts=4
